;==============================================================================

; file: "_nonstd.scm"

; Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved.

;==============================================================================

(##include "header.scm")

;==============================================================================

; Implementation of exceptions.

(implement-library-type-error-exception)

;------------------------------------------------------------------------------

; Define type checking procedures.

(define-fail-check-type will 'will)
(define-fail-check-type box 'box)

;------------------------------------------------------------------------------

; Non-standard procedures

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (define-runtime-macro pattern . rest)

  (define (form-size parms) ; this definition must match the one in "_eval.scm"
    (let loop ((l parms) (n 1))
      (if (pair? l)
        (loop (cdr l) (+ n 1))
        (if (null? l) n (- 0 n)))))

  `(##top-cte-add-macro!
     ##interaction-cte
     ',(car pattern)
     (##cons ',(form-size (cdr pattern))
             (lambda ,(cdr pattern) ,@rest))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (exit #!optional (status (macro-absent-obj)))
  (if (##eq? status (macro-absent-obj))
    (##exit)
    (macro-force-vars (status)
      (##exit status))))

(define-prim (error message . arguments)
  (macro-raise (macro-make-error-exception message arguments)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (include filename)
  `(##include ,filename))

;*********************** bug: (define-macro (make-improper z) `(define ,z '(1 . 2)))
(define-runtime-macro (define-macro pattern body . rest)
  `(##define-macro ,pattern ,body ,@rest))

(define-runtime-macro (declare . rest)
  `(##declare ,@rest))

(define-runtime-macro (namespace . rest)
  `(##namespace ,@rest))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (parameterize bindings body . rest)
  (##parameterize-build
   (##cons 'parameterize (##cons bindings (##cons body rest)))))

(define-prim (##parameterize-build src)

  (define (build src bindings body rev-params-vals)
    (cond ((##pair? bindings)
           (let ((binding (##car bindings)))
             (##shape src (##sourcify binding src) 2)
             (let* ((param (##car binding))
                    (val (##cadr binding)))
               (build src
                      (##cdr bindings)
                      body
                      (##cons (##cons (##cons (##gensym) param)
                                      (##cons (##gensym) val))
                              rev-params-vals)))))
          ((##null? bindings)
           (if (##null? rev-params-vals)
             (##cons 'let (##cons '() body))
             (let ((params-vals (##reverse rev-params-vals)))

               (define (bind params-vals)
                 (if (##null? params-vals)
                   (##cons 'let (##cons '() body))
                   (let* ((param-val (##car params-vals))
                          (param (##car param-val))
                          (val (##cdr param-val)))
                     (##list '##parameterize
                             (##car param)
                             (##car val)
                             (##list 'lambda
                                     '()
                                     (bind (##cdr params-vals)))))))

               (##list 'let
                       (let loop ((lst rev-params-vals) (bs '()))
                         (if (##null? lst)
                           bs
                           (let* ((param-val (##car lst))
                                  (param (##car param-val))
                                  (val (##cdr param-val)))
                             (loop (##cdr lst)
                                   (##cons (##list (##car param)
                                                   (##cdr param))
                                           (##cons (##list (##car val)
                                                           (##cdr val))
                                                   bs))))))
                       (##list 'if
                               (##cons 'and
                                       (##map (lambda (param-val)
                                                (let ((param
                                                       (##car param-val)))
                                                  (##list '##parameter?
                                                          (##car param))))
                                              params-vals))
                               (bind params-vals)
                               '(##raise-type-exception-parameter))))))
          (else
           (##raise-expression-parsing-exception
            'ill-formed-binding-list
            src))))

  (build (##sourcify src (##make-source #f #f))
         (##cadr src)
         (##cddr src)
         '()))

(define-prim (##raise-type-exception-parameter)
  (error "PARAMETER expected"));;;;;;;;;;;;;;;;;;;;;

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (cond-expand first-clause . other-clauses)
  (##cond-expand-build
   (##cons 'cond-expand (##cons first-clause other-clauses))))

(define-prim (##cond-expand-build src)

  (define (satisfied? src feature-requirement)
    (cond ((##symbol? feature-requirement)
           (if (##member feature-requirement ##cond-expand-features)
             #t
             #f))
          ((##pair? feature-requirement)
           (let ((first (##car feature-requirement)))
             (cond ((##eq? first 'not)
                    (##shape src (##sourcify feature-requirement src) 2)
                    (##not (satisfied? src (##cadr feature-requirement))))
                   ((or (##eq? first 'and) (##eq? first 'or))
                    (##shape src (##sourcify feature-requirement src) -1)
                    (let loop ((lst (##cdr feature-requirement)))
                      (if (##pair? lst)
                        (let ((x (##car lst)))
                          (if (##eq? (satisfied? src x) (##eq? first 'and))
                            (loop (##cdr lst))
                            (##not (##eq? first 'and))))
                        (##eq? first 'and))))
                   (else
                    (macro-raise
                     (macro-make-expression-parsing-exception
                      'ill-formed-cond-expand
                      src
                      '()))))))
          (else
           (macro-raise
            (macro-make-expression-parsing-exception
             'ill-formed-cond-expand
             src
             '())))))

  (define (build src clauses)
    (if (##pair? clauses)
      (let ((clause (##car clauses)))
        (##shape src (##sourcify clause src) -1)
        (let ((feature-requirement (##car clause)))
          (if (or (and (##eq? feature-requirement 'else)
                       (##null? (##cdr clauses)))
                  (satisfied? src feature-requirement))
            (##cons 'begin (##cdr clause))
            (build src (##cdr clauses)))))
      (macro-raise
       (macro-make-expression-parsing-exception
        'unfulfilled-cond-expand
        src
        '()))))

  (build (##sourcify src (##make-source #f #f))
         (##cdr src)))

(define ##cond-expand-features #f)
(set! ##cond-expand-features
  '(
    srfi-0 SRFI-0
    srfi-4 SRFI-4
    srfi-6 SRFI-6
    srfi-8 SRFI-8
    srfi-9 SRFI-9
    srfi-18 SRFI-18
    srfi-21 SRFI-21
    srfi-22 SRFI-22
    srfi-23 SRFI-23
    srfi-27 SRFI-27
   ))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (receive formals expression first-body . other-body)
  `(##call-with-values
    (lambda () ,expression)
    (lambda ,formals ,first-body ,@other-body)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (##type-field-count type)
  (if type
    (let ((fields (##type-fields type)))
      (##fixnum.+ (##type-field-count (##type-super type))
                  (##fixnum.quotient (##vector-length fields) 3)))
    0))

(define-prim (##type-all-fields type)
  (if type
    (let ((fields (##type-fields type)))
      (##append (##type-all-fields (##type-super type))
                (##vector->list fields)))
    '()))

(define-prim (##type-all-cfields type)
  (if type
    (let ((cfields (##type-cfields type)))
      (##append (##type-all-cfields (##type-super type))
                (##vector->list cfields)))
    '()))

(define-prim (##define-type-expand
              form-name
              super-type-static
              super-type-dynamic
              args)

  (define (generate
           name
           flags
           id
           extender
           constructor
           constant-constructor
           predicate
           implementer
           type-exhibitor
           prefix
           fields
           total-fields)

    (define (generate-fields)
      (let loop ((lst1 (##reverse fields))
                 (lst2 '()))
        (if (##pair? lst1)
          (let* ((field
                  (##car lst1))
                 (descr
                  (##cdr field))
                 (field-name
                  (##vector-ref descr 0))
                 (options
                  (##vector-ref descr 4))
                 (attributes
                  (##vector-ref descr 5))
                 (init
                  (cond ((##assq 'init: attributes)
                         =>
                         (lambda (x) (##constant-expression-value (##cdr x))))
                        (else
                         #f))))
            (loop (##cdr lst1)
                  (##cons field-name
                          (##cons options
                                  (##cons init
                                          lst2)))))
          (##list->vector lst2))))

    (define (pos-in-list x lst)
      (let loop ((lst lst)
                 (i 0))
        (cond ((##not (##pair? lst))
               #f)
              ((##fixnum.= (##car lst) x)
               i)
              (else
               (loop (##cdr lst)
                     (##fixnum.+ i 1))))))

    (define (generate-parameter i)
      (##string->symbol
       (##string-append "param"
                        (##number->string i 10))))

    (define (generate-parameters cfields)
      (let loop ((i (##length cfields))
                 (params '()))
        (if (##fixnum.< i 1)
          params
          (loop (##fixnum.- i 1)
                (##cons (generate-parameter i)
                        params)))))

    (define (generate-initializations fields cfields)
      (let loop ((i 1)
                 (lst fields)
                 (rev-inits '()))
        (if (##pair? lst)
          (let ((rest (##cddr lst)))
            (loop (##fixnum.+ i 1)
                  (##cdr rest)
                  (##cons
                   (let ((x (pos-in-list i cfields)))
                     (if x
                       (generate-parameter (##fixnum.+ x 1))
                       (##list 'quote (##list 'quote (##car rest)))))
                   rev-inits)))
          (##reverse rev-inits))))

    (let* ((macros?
            (##not (##fixnum.= (##fixnum.bitwise-and flags 4) 0)))
           (generative?
            (##not id))
           (augmented-id-str
            (##string-append
             "##type-"
             (##number->string total-fields 10)
             "-"
             (##symbol->string (if generative? name id))))
           (type-fields
            (generate-fields))
           (type-cfields
            (##list->vector
             (if constructor
               (let* ((field-names
                       (##map ##car fields))
                      (start
                       (##fixnum.+ (##type-field-count super-type-static) 1)))
                 (##map (lambda (x)
                          (##fixnum.+ (pos-in-list x field-names) start))
                        (##cdr constructor)))
               '())))
           (type-static
            (##structure
             ##type-type
             (if generative?
               (##make-uninterned-symbol augmented-id-str)
               (##string->symbol augmented-id-str))
             name
             flags
             super-type-static
             type-fields
             type-cfields))
           (type-expression
            (if generative?
              (##string->symbol augmented-id-str)
              `',type-static))
           (type-id-expression
            (if generative?
              `(let ()
                 (##declare (extended-bindings) (not safe))
                 (##type-id ,type-expression))
              `',(##type-id type-static)))
           (all-fields
            (##type-all-fields type-static))
           (all-cfields
            (##type-all-cfields type-static))
           (parameters
            (generate-parameters all-cfields))
           (initializations
            (generate-initializations all-fields all-cfields)))

      (define (generate-getter-and-setter field tail)
        (let* ((descr
                (##cdr field))
               (field-name
                (##vector-ref descr 0))
               (field-index
                (##vector-ref descr 1))
               (getter
                (##vector-ref descr 2))
               (setter
                (##vector-ref descr 3))
               (getter-def
                (if getter
                  (let ((getter-name
                         (if (##eq? getter #t)
                           (##symbol-append prefix
                                            name
                                            '-
                                            field-name)
                           getter)))
                    (if macros?
                      `((##define-macro (,getter-name obj)
                          (##list '(let ()
                                     (##declare (extended-bindings))
                                     ##structure-ref)
                                  obj
                                  ,field-index
                                  ',type-expression
                                  #f)))
                      `((define (,getter-name obj)
                          ((let ()
                             (##declare (extended-bindings))
                             ##structure-ref)
                           obj
                           ,field-index
                           ,type-expression
                           ,getter-name)))))
                  `()))
               (setter-def
                (if setter
                  (let ((setter-name
                         (if (##eq? setter #t)
                           (##symbol-append prefix
                                            name
                                            '-
                                            field-name
                                            '-set!)
                           setter)))
                    (if macros?
                      `((##define-macro (,setter-name obj val)
                          (##list '(let ()
                                     (##declare (extended-bindings))
                                     ##structure-set!)
                                  obj
                                  val
                                  ,field-index
                                  ',type-expression
                                  #f)))
                      `((define (,setter-name obj val)
                          ((let ()
                             (##declare (extended-bindings))
                             ##structure-set!)
                           obj
                           val
                           ,field-index
                           ,type-expression
                           ,setter-name)))))
                  `())))
          (##append getter-def (##append setter-def tail))))

      (define (generate-structure-type-definition)
        `(define ,type-expression
           ((let ()
              (##declare (extended-bindings))
              ##structure)
            ',##type-type
            ((let ()
               (##declare (extended-bindings))
               ##make-uninterned-symbol)
             ,augmented-id-str)
            ',name
            ',(##type-flags type-static)
            ,super-type-dynamic
            ',(##type-fields type-static)
            ',(##type-cfields type-static))))

      (define (generate-constructor-predicate-getters-setters)
        `(,@(if type-exhibitor
              (if macros?
                `((##define-macro (,type-exhibitor)
                    ',type-expression))
                `((define (,type-exhibitor)
                    ,type-expression)))
              '())

          ,@(if constructor
              (if macros?
                `((##define-macro (,(##car constructor) ,@parameters)
                    (##list '(let ()
                               (##declare (extended-bindings))
                               ##structure)
                            ',type-expression
                            ,@initializations)))
                `((define (,(##car constructor) ,@parameters)
                    (##declare (extended-bindings))
                    (##structure
                     ,type-expression
                     ,@initializations))))
              '())

          ,@(if constant-constructor
              `((##define-macro (,constant-constructor ,@parameters)
                  (##define-type-construct-constant
                   ',constant-constructor
                   ,type-expression
                   ,@initializations)))
              '())

          ,@(if predicate
              (if macros?
                `((##define-macro (,predicate obj)
                    (##list ,(if extender
                               ''(let ()
                                   (##declare (extended-bindings))
                                   ##structure-instance-of?)
                               ''(let ()
                                   (##declare (extended-bindings))
                                   ##structure-direct-instance-of?))
                            obj
                            ',type-id-expression)))
                `((define (,predicate obj)
                    (##declare (extended-bindings))
                    (,(if extender
                        '##structure-instance-of?
                        '##structure-direct-instance-of?)
                     obj
                     ,type-id-expression)))))

          ,@(let loop ((lst1 (##reverse fields))
                       (lst2 '()))
              (if (##pair? lst1)
                (loop (##cdr lst1)
                      (generate-getter-and-setter (##car lst1) lst2))
                lst2))))

      (define (generate-definitions)
        (if generative?
          (##cons (generate-structure-type-definition)
                  (generate-constructor-predicate-getters-setters))
          (generate-constructor-predicate-getters-setters)))

      `(begin

         ,@(if extender
             (##list `(##define-macro (,extender . args)
                        (##define-type-expand
                         ',extender
                         ',type-static
                         ,(if generative? `',type-expression #f)
                         args)))
             '())

         ,@(if implementer
             (if macros?
               (##cons `(##define-macro (,implementer)
                          ',(if generative?
                              (generate-structure-type-definition)
                              '(begin)))
                       (generate-constructor-predicate-getters-setters))
               (##list `(##define-macro (,implementer)
                          ',(##cons 'begin
                                    (generate-definitions)))))
             (generate-definitions)))))

(let ((q
  (##define-type-parser
   form-name
   super-type-static
   super-type-dynamic
   args
   generate)))
;(pp q)
 q));;;;;;;;;;;;;;;;;;;;;;

(define-prim (##define-type-parser
              form-name
              super-type-static
              super-type-dynamic
              args
              cont)

  (define (err)
    (##ill-formed-special-form form-name args))

  (define (parse-attributes name rest)
    (let loop1 ((lst rest)
                (field-index (##type-field-count super-type-static))
                (options 0)
                (flags '())
                (rev-fields '()))

      (define allowed-field-options
        '((printable:     . (-2 . 0))
          (unprintable:   . (-2 . 1))
          (read-write:    . (-3 . 0))
          (read-only:     . (-3 . 2))
          (equality-test: . (-5 . 0))
          (equality-skip: . (-5 . 4))))

      (define (update-options options opt)
        (let* ((x (##cdr opt))
               (m (##car x))
               (b (##cdr x)))
          (##fixnum.bitwise-ior (##fixnum.bitwise-and options m) b)))

      (define (parse-field-attributes
               field-name
               getter
               setter
               local-options
               rest)
        (let loop2 ((lst2 rest)
                    (local-options local-options)
                    (attributes '()))
          (cond ((##pair? lst2)
                 (let ((attribute (##car lst2)))
                   (cond ((##member attribute
                                    '(init:))
                          (let ((rest (##cdr lst2)))
                            (if (and (##pair? rest)
                                     (##not (##assq attribute attributes)))
                              (let ((val (##car rest)))
                                (if (##constant-expression? val)
                                  (loop2 (##cdr rest)
                                         local-options
                                         (##cons (##cons attribute val)
                                                 attributes))
                                  (err)))
                              (err))))
                         ((##assq attribute
                                  allowed-field-options)
                          =>
                          (lambda (opt)
                            (loop2 (##cdr lst2)
                                   (update-options local-options opt)
                                   attributes)))
                         (else
                          (err)))))
                ((##null? lst2)
                 (let ((read-only?
                        (##not
                         (##fixnum.= (##fixnum.bitwise-and local-options 2)
                                     0))))
                   (if (and (##symbol? setter)
                            read-only?)
                     (err)
                     (loop1 (##cdr lst)
                            (##fixnum.+ field-index 1)
                            options
                            flags
                            (##cons (##cons field-name
                                            (##vector
                                             field-name
                                             (##fixnum.+ field-index 1)
                                             getter
                                             (if read-only? #f setter)
                                             local-options
                                             attributes))
                                    rev-fields)))))
                (else
                 (err)))))

      (cond ((##pair? lst)
             (let ((next (##car lst)))
               (cond ((##symbol? next)
                      (if (##not (##assq next rev-fields))
                        (parse-field-attributes
                         next
                         #t
                         #t
                         options
                         '())
                        (err)))
                     ((##pair? next)
                      (let* ((field-name (##car next))
                             (rest (##cdr next)))
                        (if (and (##symbol? field-name)
                                 (##not (##assq field-name rev-fields)))
                          (if (##pair? rest)
                            (let ((getter (##car rest)))
                              (if (##symbol? getter)
                                (let ((rest (##cdr rest)))
                                  (if (##pair? rest)
                                    (let ((setter (##car rest)))
                                      (if (##symbol? setter)
                                        (parse-field-attributes
                                         field-name
                                         getter
                                         setter
                                         (##fixnum.bitwise-and options -3)
                                         (##cdr rest))
                                        (parse-field-attributes
                                         field-name
                                         getter
                                         #f
                                         (##fixnum.bitwise-ior options 2)
                                         rest)))
                                    (parse-field-attributes
                                     field-name
                                     getter
                                     #f
                                     (##fixnum.bitwise-ior options 2)
                                     rest)))
                                (parse-field-attributes
                                 field-name
                                 #t
                                 #t
                                 options
                                 rest)))
                            (parse-field-attributes
                             field-name
                             #t
                             #t
                             options
                             rest))
                          (err))))
                     ((##member next
                                '(id:
                                  constructor:
                                  constant-constructor:
                                  predicate:
                                  extender:
                                  implementer:
                                  type-exhibitor:
                                  prefix:))
                      (let ((rest (##cdr lst)))
                        (if (and (##pair? rest)
                                 (##not (##assq next flags)))
                          (let ((val (##car rest)))
                            (if (cond ((##eq? next 'constructor:)
                                       (if (##pair? val)
                                         (if (##symbol? (##car val))
                                           (let loop ((lst1 (##cdr val))
                                                      (lst2 '()))
                                             (if (##pair? lst1)
                                               (let ((x (##car lst1)))
                                                 (if (and (##symbol? x)
                                                          (##not (##member
                                                                  x
                                                                  lst2)))
                                                   (loop (##cdr lst1)
                                                         (##cons x lst2))
                                                   #f))
                                               (##null? lst1)))
                                           #f)
                                         (or (##not val)
                                             (##symbol? val))))
                                      (else
                                       (or (##symbol? val)
                                           (and (##memq
                                                 next
                                                 '(predicate:
                                                   constant-constructor:))
                                                (##not val)))))
                              (loop1 (##cdr rest)
                                     field-index
                                     options
                                     (##cons (##cons next val) flags)
                                     rev-fields)
                              (err)))
                          (err))))
                     ((##member next
                                '(opaque:
                                  macros:))
                      (if (##not (##assq next flags))
                        (loop1 (##cdr lst)
                               field-index
                               options
                               (##cons (##cons next #t) flags)
                               rev-fields)
                        (err)))
                     ((##assq next
                              allowed-field-options)
                      =>
                      (lambda (opt)
                        (loop1 (##cdr lst)
                               field-index
                               (update-options options opt)
                               flags
                               rev-fields)))
                     (else
                      (err)))))
            ((##null? lst)
             (let* ((fields
                     (##reverse rev-fields))
                    (prefix
                     (cond ((##assq 'prefix: flags)
                            =>
                            ##cdr)
                           (else
                            '||)))
                    (id
                     (cond ((##assq 'id: flags)
                            =>
                            ##cdr)
                           (else
                            #f)))
                    (extender
                     (cond ((##assq 'extender: flags)
                            =>
                            ##cdr)
                           (else
                            #f)))
                    (constructor
                     (cond ((##assq 'constructor: flags)
                            =>
                            (lambda (x)
                              (let ((val (##cdr x)))
                                (cond ((##not val)
                                       #f)
                                      ((##symbol? val)
                                       (##cons val (##map ##car fields)))
                                      (else
                                       val)))))
                           (else
                            (##cons (##symbol-append prefix
                                                     'make-
                                                     name)
                                    (##map ##car fields)))))
                    (constant-constructor
                     (cond ((##assq 'constant-constructor: flags)
                            =>
                            ##cdr)
                           ((or (##not constructor)
                                (##not id))
                            #f)
                           (else
                            (##symbol-append prefix
                                             'make-constant-
                                             name))))
                    (predicate
                     (cond ((##assq 'predicate: flags)
                            =>
                            ##cdr)
                           (else
                            (##symbol-append prefix
                                             name
                                             '?))))
                    (implementer
                     (cond ((##assq 'implementer: flags)
                            =>
                            ##cdr)
                           (else
                            #f)))
                    (type-exhibitor
                     (cond ((##assq 'type-exhibitor: flags)
                            =>
                            ##cdr)
                           (else
                            #f))))
               (if (and constant-constructor
                        (or (##not constructor)
                            (##not id)))
                 (err)
                 (cont
                  name
                  (##fixnum.+ (if (##assq 'opaque: flags) 1 0)
                              (if extender 2 0)
                              (if (##assq 'macros: flags) 4 0)
                              (if constructor 8 0))
                  id
                  extender
                  constructor
                  constant-constructor
                  predicate
                  implementer
                  type-exhibitor
                  prefix
                  fields
                  field-index))))
            (else
             (err)))))

  (if (##pair? args)
    (let* ((name (##car args))
           (rest (##cdr args)))
      (if (##symbol? name)
        (parse-attributes name rest)
        (err)))
    (err)))

(define-prim (##define-type-construct-constant form-name type . fields)
  (let loop ((lst1 fields)
             (lst2 '()))
    (if (##pair? lst1)
      (let ((field (##car lst1)))
        (if (##constant-expression? field)
          (loop (##cdr lst1)
                (##cons (##constant-expression-value field)
                        lst2))
          (##ill-formed-special-form form-name fields)))
      `',(##apply ##structure
                  (##cons type (##reverse lst2))))))

(define-prim (##ill-formed-special-form form-name args)
  (##raise-expression-parsing-exception
   'ill-formed-special-form
   (##sourcify
    (##cons form-name args)
    (##make-source #f #f))
   form-name))

(define-prim (##constant-expression? expr)
  (or (##self-eval? expr)
      (and (##pair? expr)
           (##eq? (##car expr) 'quote)
           (let ((rest (##cdr expr)))
             (and (##pair? rest)
                  (##null? (##cdr rest)))))))

(define-prim (##constant-expression-value expr)
  (if (##self-eval? expr)
    expr
    (##cadr expr)))

(define-prim (##symbol-append . symbols)
  (##string->symbol
   (##apply ##string-append
            (##map ##symbol->string symbols))))

(define-runtime-macro (define-type . args)
  (##define-type-expand 'define-type #f #f args))

(define-runtime-macro (define-structure . args)
  (##define-type-expand 'define-structure #f #f args))

(define-runtime-macro (define-record-type name constructor predicate . fields)
  `(define-type ,name
     constructor: ,constructor
     predicate: ,predicate
     ,@fields))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (set-gc-report! report?);;;;;;;;;;;;;;;;;;;;;;
  (##set-gc-report! report?))

(define ##gc-report? #f)

(define-prim (##set-gc-report! report?)
  (##declare (not interrupts-enabled))
  (set! ##gc-report? (if report? #t #f)))

(define-prim (##display-gc-report)
  (let ((port ##stderr-port))
    (if (and port
             (let* ((settings
                     (##set-debug-settings! 0 0))
                    (level
                     (macro-debug-settings-level settings)))
               (or (##fixnum.< 1 level)
                   ##gc-report?)))
      (let* ((stats
              (##process-statistics))
             (bytes-allocated
              (##f64vector-ref stats 7))
             (last-gc-real-time
              (##f64vector-ref stats 12))
             (last-gc-heap-size
              (##f64vector-ref stats 13))
             (last-gc-live
              (##f64vector-ref stats 14))
             (last-gc-movable
              (##f64vector-ref stats 15))
             (last-gc-nonmovable
              (##f64vector-ref stats 16)))

        (define (scale x m)
          (##flonum.->exact-int (##flonum.round (##flonum.* x m))))

        (##write-string "*** GC: " port)
        (##write (scale last-gc-real-time 1000.0) port)
        (##write-string " ms, " port)
        (##write (scale bytes-allocated 9.765625e-4) port)
        (##write-string "K alloc, " port)
        (##write (scale last-gc-heap-size 9.765625e-4) port)
        (##write-string "K heap, " port)
        (##write (scale last-gc-live 9.765625e-4) port)
        (##write-string "K live (" port)
        (##write (scale (##flonum./ last-gc-live last-gc-heap-size) 100.0) port)
        (##write-string "% " port)
        (##write (##flonum.->exact-int last-gc-movable) port)
        (##write-string "+" port)
        (##write (##flonum.->exact-int last-gc-nonmovable) port)
        (##write-string ")" port)
        (##newline port)))))

(##add-gc-interrupt-job! ##display-gc-report)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (##void))

(define-prim (void)
  (##void))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (process-times)
  (##process-times))

(define-prim (cpu-time)
  (let ((v (##process-times)))
    (##+ (##f64vector-ref v 0) (##f64vector-ref v 1))))

(define-prim (real-time)
  (let ((v (##process-times)))
    (##f64vector-ref v 2)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (time expr)
  `(##time (lambda () ,expr) ',expr))

(define-prim (##time thunk expr)
  (let ((at-start (##process-statistics)))
    (let ((result (thunk)))
      (let ((at-end (##process-statistics)))

        (define (secs->msecs x)
          (##inexact->exact (##round (##* x 1000))))

        (let* ((out
                (##repl-out))
               (user-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 0)
                      (##f64vector-ref at-start 0))))
               (sys-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 1)
                      (##f64vector-ref at-start 1))))
               (cpu-time
                (##+ user-time sys-time))
               (real-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 2)
                      (##f64vector-ref at-start 2))))
               (gc-user-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 3)
                      (##f64vector-ref at-start 3))))
               (gc-sys-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 4)
                      (##f64vector-ref at-start 4))))
               (gc-real-time
                (secs->msecs
                 (##- (##f64vector-ref at-end 5)
                      (##f64vector-ref at-start 5))))
               (nb-gcs
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 6)
                      (##f64vector-ref at-start 6))))
               (minflt
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 10)
                      (##f64vector-ref at-start 10))))
               (majflt
                (##flonum.->exact-int
                 (##- (##f64vector-ref at-end 11)
                      (##f64vector-ref at-start 11))))
               (bytes-allocated
                (##flonum.->exact-int
                 (##- (##- (##f64vector-ref at-end 7)
                           (##f64vector-ref at-start 7))
                      (##+ (if (##interp-procedure? thunk)
                             (##f64vector-ref at-end 8) ; thunk call frame space
                             (macro-inexact-+0))
                           (##f64vector-ref at-end 9)))))) ; at-end structure space

          (define (pluralize n msg)
            (##write-string "    " out)
            (if (##= n 0)
              (##write-string "no" out)
              (##write n out))
            (##write-string msg out)
            (if (##not (##= n 1))
              (##write-string "s" out)))

          (##write (##list 'time expr) out)
          (##newline out)

          (##write-string "    " out)
          (##write real-time out)
          (##write-string " ms real time" out)
          (##newline out)

          (##write-string "    " out)
          (##write cpu-time out)
          (##write-string " ms cpu time (" out)
          (##write user-time out)
          (##write-string " user, " out)
          (##write sys-time out)
          (##write-string " system)" out)
          (##newline out)

          (pluralize nb-gcs " collection")
          (if (##not (##= nb-gcs 0))
            (begin
              (##write-string " accounting for " out)
              (##write gc-real-time out)
              (##write-string " ms real time (" out)
              (##write gc-user-time out)
              (##write-string " user, " out)
              (##write gc-sys-time out)
              (##write-string " system)" out)))
          (##newline out)

          (pluralize bytes-allocated " byte")
          (##write-string " allocated" out)
          (##newline out)

          (pluralize minflt " minor fault")
          (##newline out)

          (pluralize majflt " major fault")
          (##newline out)

          result)))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define ##gensym-counter -1)

(define-prim (##gensym #!optional (p (macro-absent-obj)))
  (let ((prefix
         (if (##eq? p (macro-absent-obj))
           'g
           p)))
    (macro-check-symbol prefix 1 (gensym p)
      (let ((new-count
             (##fixnum.modulo
              (##fixnum.+ ##gensym-counter 1)
              1000000)))
        ; Note: it is unimportant if the increment of ##gensym-counter
        ; is not atomic; it simply means a possible close repetition
        ; of the same name
        (set! ##gensym-counter new-count)
        (##make-uninterned-symbol
         (if (##eq? prefix 'g)
           new-count ; ##symbol->string will create the string
           (##string-append (##symbol->string prefix)
                            (##number->string new-count 10))))))))

(define-prim (gensym #!optional (p (macro-absent-obj)))
  (macro-force-vars (p)
    (##gensym p)))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(##define-macro (macro-will-size) 3)

(define-prim (##will? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-willtable))
       (##fixnum.= (##vector-length obj) (macro-will-size))))

(define-prim (will? x)
  (macro-force-vars (x)
    (##will? x)))

(define-prim (##make-will testator action)
  (macro-make-will testator action))

(define-prim (make-will testator action)
  (macro-force-vars (action)
    (macro-check-procedure action 2 (make-will testator action)
      (macro-make-will testator action))))

(define-prim (##will-testator will)
  (macro-will-testator will))

(define-prim (will-testator will)
  (macro-force-vars (will)
    (macro-check-will will 1 (will-testator will)
      (macro-will-testator will))))

(define-prim (##will-execute! will)
  (macro-will-execute! will))

(define-prim (will-execute! will)
  (macro-force-vars (will)
    (macro-check-will will 1 (will-execute! will)
      (macro-will-execute! will))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (##box? obj)
  (and (##subtyped? obj)
       (##eq? (##subtype obj) (macro-subtype-boxvalues))
       (##fixnum.= (##vector-length obj) 1)))

(define-prim (box? obj)
  (macro-force-vars (obj)
    (##box? obj)))

(define-prim (##box obj)
  (##subtype-set! (##vector obj) (macro-subtype-boxvalues)))

(define-prim (box obj)
  (##box obj))

(define-prim (##unbox box)
  (##vector-ref box 0))

(define-prim (unbox box)
  (macro-force-vars (box)
    (macro-check-box box 1 (unbox box)
      (##unbox box))))

(define-prim (##set-box! box val)
  (##vector-set! box 0 val))

(define-prim (set-box! box val)
  (macro-force-vars (box)
    (macro-check-box box 1 (set-box! box val)
      (##set-box! box val))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (command-line)
  ##processed-command-line)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-prim (getenv name)
  (macro-force-vars (name)
    (macro-check-string name 1 (getenv name)
      (##getenv name))))

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

(define-runtime-macro (infix.!x x)
  `(not ,x))

(define-runtime-macro (infix.++x x)
  (##infix-update-in-place 'infix.++x x 'infix.x+y 1 #t))

(define-runtime-macro (infix.x++ x)
  (##infix-update-in-place 'infix.x++ x 'infix.x+y 1 #f))

(define-runtime-macro (infix.--x x)
  (##infix-update-in-place 'infix.--x x 'infix.x-y 1 #t))

(define-runtime-macro (infix.x-- x)
  (##infix-update-in-place 'infix.x-- x 'infix.x-y 1 #f))

(define-runtime-macro (infix.~x x)
  `(bitwise-not ,x))

(define-runtime-macro (infix.x%y x y)
  `(modulo ,x ,y))

(define-runtime-macro (infix.x*y x y)
  `(* ,x ,y))

(define-runtime-macro (infix.*x x)
  (##infix-lvalue-fetch (##list 'infix.*x x)))

(define-runtime-macro (infix.x/y x y)
  `(/ ,x ,y))

(define-runtime-macro (infix.x+y x y)
  `(+ ,x ,y))

(define-runtime-macro (infix.+x x)
  `(+ ,x))

(define-runtime-macro (infix.x-y x y)
  `(- ,x ,y))

(define-runtime-macro (infix.-x x)
  `(- ,x))

(define-runtime-macro (infix.x<<y x y)
  `(arithmetic-shift ,x ,y))

(define-runtime-macro (infix.x>>y x y)
  `(arithmetic-shift ,x (- ,y)))

(define-runtime-macro (infix.x<y x y)
  `(< ,x ,y))

(define-runtime-macro (infix.x<=y x y)
  `(<= ,x ,y))

(define-runtime-macro (infix.x>y x y)
  `(> ,x ,y))

(define-runtime-macro (infix.x>=y x y)
  `(>= ,x ,y))

(define-runtime-macro (infix.x!=y x y)
  `(not (equal? ,x ,y)))

(define-runtime-macro (infix.x==y x y)
  `(equal? ,x ,y))

(define-runtime-macro (infix.x&y x y)
  `(bitwise-and ,x ,y))

(define-prim (##infix-id x)
  (if (##pair? x)
    (let* ((first (##car x))
           (rest (##cdr x)))
      (if (and (or (##eq? first 'infix.identifier)
                   (##eq? first 'infix.prefix))
               (##pair? rest))
        (let* ((second (##car rest))
               (rest (##cdr rest)))
          (if (and (##symbol? second)
                   (##null? rest))
            second
            #f))
        #f))
    #f))
  
(define-runtime-macro (infix.&x x)
  (##infix-lvalue-access
   'infix.&x
   (##list x)
   x
   (lambda (prepare fetch only-fetch store only-store)
     (let* ((set (##gensym))
            (val (##gensym)))
       (prepare
        `(lambda ,set
           (if (##pair? ,set)
             (let ((,val (##car ,set)))
               ,(store val)
               ,val)
             ,(fetch))))))))

(define-prim (##infix-lvalue-access form-name args x cont)

  (define (err)
    (##ill-formed-special-form form-name args))

  (if (##pair? x)
    (let* ((first (##car x))
           (rest (##cdr x)))
      (if (##pair? rest)
        (let* ((second (##car rest))
               (rest (##cdr rest)))
          (cond ((##pair? rest)
                 (let* ((third (##car rest))
                        (rest (##cdr rest)))
                   (cond ((##not (##null? rest))
                          (err))
                         ((##eq? first 'infix.index)
                          (let* ((vect (##gensym))
                                 (index (##gensym)))
                            (cont (lambda (body)
                                    `(let ((,vect ,second) (,index ,third))
                                       ,body))
                                  (lambda ()
                                    `(vector-ref ,vect ,index))
                                  (lambda ()
                                    `(vector-ref ,second ,third))
                                  (lambda (val)
                                    `(vector-set! ,vect ,index ,val))
                                  (lambda (val)
                                    `(vector-set! ,second ,third ,val)))))
                         ((and (or (##eq? first 'infix.arrow)
                                   (##eq? first 'infix.dot))
                               (##infix-id third))
                          =>
                          (lambda (id)
                            (let* ((struct (##gensym))
                                   (mutator
                                    (##string->symbol
                                     (##string-append
                                      (##symbol->string id)
                                      "-set!"))))
                            (cont (lambda (body)
                                    `(let ((,struct ,second))
                                       ,body))
                                  (lambda ()
                                    `(,id ,struct))
                                  (lambda ()
                                    `(,id ,second))
                                  (lambda (val)
                                    `(,mutator ,struct ,val))
                                  (lambda (val)
                                    `(,mutator ,second ,val))))))
                         (else
                          (err)))))
                ((##null? rest)
                 (cond ((##eq? first 'infix.*x)
                        (let ((ptr (##gensym)))
                          (cont (lambda (body)
                                  `(let ((,ptr ,second))
                                     ,body))
                                (lambda ()
                                  `(,ptr))
                                (lambda ()
                                  `(,second))
                                (lambda (val)
                                  `(,ptr ,val))
                                (lambda (val)
                                  `(,second ,val)))))
                       ((and (or (##eq? first 'infix.identifier)
                                 (##eq? first 'infix.prefix))
                             (##symbol? second))
                        (let ((var (##gensym)))
                          (cont (lambda (body)
                                  body)
                                (lambda ()
                                  second)
                                (lambda ()
                                  second)
                                (lambda (val)
                                  `(set! ,second ,val))
                                (lambda (val)
                                  `(set! ,second ,val)))))
                       (else
                        (err))))
                (else
                 (err))))
        (err)))
    (err)))

(define-prim (##infix-lvalue-fetch form)
  (##infix-lvalue-access
   (##car form)
   (##cdr form)
   form
   (lambda (prepare fetch only-fetch store only-store)
     (only-fetch))))

(define-prim (##infix-update-in-place form-name x operator operand2 pre?)
  (##infix-lvalue-access
   form-name
   (if (##eq? operand2 1)
     (##list x)
     (##list x operand2))
   x
   (lambda (prepare fetch only-fetch store only-store)
     (let ((val (##gensym)))
       (prepare
        (if pre?
          `(let ((,val (,operator ,(fetch) ,operand2)))
             ,(store val)
             ,val)
          `(let ((,val ,(fetch)))
             ,(store `(,operator ,val ,operand2))
             ,val)))))))

(define-runtime-macro (infix.x^y x y)
  `(bitwise-xor ,x ,y))

(define-runtime-macro (|infix.x\|y| x y)
  `(bitwise-ior ,x ,y))

(define-runtime-macro (infix.x&&y x y)
  `(and ,x ,y))

(define-runtime-macro (|infix.x\|\|y| x y)
  `(or ,x ,y))

(define-runtime-macro (infix.x?y:z x y z)
  `(if ,x ,y ,z))

(define-runtime-macro (infix.x:y x y)
  `(cons ,x ,y))

(define-runtime-macro (infix.x%=y x y)
  (##infix-update-in-place 'infix.x%=y x 'infix.x%y y #t))

(define-runtime-macro (infix.x&=y x y)
  (##infix-update-in-place 'infix.x&=y x 'infix.x&y y #t))

(define-runtime-macro (infix.x*=y x y)
  (##infix-update-in-place 'infix.x*=y x 'infix.x*y y #t))

(define-runtime-macro (infix.x+=y x y)
  (##infix-update-in-place 'infix.x+=y x 'infix.x+y y #t))

(define-runtime-macro (infix.x-=y x y)
  (##infix-update-in-place 'infix.x-=y x 'infix.x-y y #t))

(define-runtime-macro (infix.x/=y x y)
  (##infix-update-in-place 'infix.x/=y x 'infix.x/y y #t))

(define-runtime-macro (infix.x<<=y x y)
  (##infix-update-in-place 'infix.x<<=y x 'infix.x<<y y #t))

(define-runtime-macro (infix.x=y x y)
  (##infix-lvalue-access
   'infix.x=y
   (##list x y)
   x
   (lambda (prepare fetch only-fetch store only-store)
     (let ((val (##gensym)))
       (prepare
        `(let ((,val ,y))
           ,(store val)
           ,val))))))

(define-runtime-macro (infix.x>>=y x y)
  (##infix-update-in-place 'infix.x>>=y x 'infix.x>>y y #t))

(define-runtime-macro (infix.x^=y x y)
  (##infix-update-in-place 'infix.x^=y x 'infix.x^y y #t))

(define-runtime-macro (|infix.x\|=y| x y)
  (##infix-update-in-place '|infix.x\|=y| x '|infix.x\|y| y #t))

(define-runtime-macro (infix.x:=y x y)
  (##infix-lvalue-access
   'infix.x:=y
   (##list x y)
   x
   (lambda (prepare fetch only-fetch store only-store)
     (let ((val (##gensym)))
       (prepare
        `(let ((,val ,y))
           ,(store val)
           ,val))))))

(define-runtime-macro (|infix.x,y| x y)
  `(begin ,x ,y))

(define-runtime-macro (infix.cons expr1 expr2)
  `(cons ,expr1 ,expr2))

(define-runtime-macro (infix.list expr1 expr2)
  `(cons ,expr1 ,expr2))

(define-runtime-macro (infix.null)
  `'())

(define-runtime-macro (infix.new identifier . args)
  (cond ((##infix-id identifier)
         =>
         (lambda (id)
           `(,(##string->symbol
               (##string-append "make-"
                                (##symbol->string id)))
             ,@args)))
        (else
         (##ill-formed-special-form 'infix.new (##cons identifier args)))))

(define-runtime-macro (infix.call func . args)
  `(,func ,@args))

(define-runtime-macro (infix.index expr1 expr2)
  (##infix-lvalue-fetch (##list 'infix.index expr1 expr2)))

(define-runtime-macro (infix.arrow expr identifier)
  (##infix-lvalue-fetch (##list 'infix.arrow expr identifier)))

(define-runtime-macro (infix.dot expr identifier)
  (##infix-lvalue-fetch (##list 'infix.dot expr identifier)))

(define-runtime-macro (infix.literal value)
  `',value)

(define-runtime-macro (infix.identifier identifier)
  (##infix-lvalue-fetch (##list 'infix.identifier identifier)))

(define-runtime-macro (infix.prefix datum)
  datum)

(define-runtime-macro (infix.if expr stat1 . stat2)
  `(if ,expr ,stat1 ,@stat2))

(define-runtime-macro (infix.while expr stat)
  (let ((loop (gensym)))
    `(let ,loop () (if ,expr (begin ,stat (,loop))))))

(define-runtime-macro (infix.do-while stat expr)
  (let ((loop (gensym)))
    `(let ,loop () (begin ,stat (if ,expr (,loop))))))

(define-runtime-macro (infix.for stat1 expr2 expr3 stat2)
  (if (##equal? stat1 '(infix.compound) #f)
    (let* ((loop (gensym))
           (body `(begin ,stat2 ,@(if expr3 `(,expr3) '()) (,loop))))
      `(let ,loop ()
         ,(if expr2
            `(if ,expr2 ,body)
            body)))
    `(infix.compound
      ,stat1
      (infix.for (infix.compound) ,expr2 ,expr3 ,stat2))))

(define-runtime-macro (infix.compound . stats)
  (##infix-compound-expand 'infix.compound stats))

(define-runtime-macro (infix.procedure-body . stats)
  (##infix-compound-expand 'infix.procedure-body stats))

(define-prim (##infix-compound-expand form-name stats)

  (define (expand lst1 lst2 first?)
    (cond ((##pair? lst1)
           (let ((stat (##car lst1)))
             (cond ((and (##pair? stat)
                         (##eq? (##car stat) 'infix.define-procedure))
                    (expand (##cdr lst1)
                            (##cons stat lst2)
                            first?))
                   ((##not (##null? lst2))
                    `((let () ,@(##reverse lst2) ,@(expand lst1 '() #t))))
                   ((##infix-variable-binding stat)
                    =>
                    (lambda (binding)
                      `((let (,binding) ,@(expand (##cdr lst1) '() #t)))))
                   ((##null? (##cdr lst1))
                    (if (and (##eq? form-name 'infix.procedure-body)
                             (##pair? stat)
                             (##eq? (##car stat) 'infix.return))
                      (let ((rest (##cdr stat)))
                        (cond ((##null? rest)
                               `((void)))
                              ((and (##pair? rest)
                                    (##null? (##cdr rest)))
                               `(,(##car rest)))
                              (else
                               `(,stat))))
                      `(,stat)))
                   (else
                    `(,stat ,@(expand (##cdr lst1) '() #f))))))
          (first?
           '((void)))
          (else
           '())))

  (if (##null? stats)
    `(void)
    `(begin ,@(expand stats '() #t))))

(define-runtime-macro (infix.define-variable identifier type dims init)
  (cond ((##infix-variable-binding
          `(infix.define-variable ,identifier ,type ,dims ,init))
         =>
         (lambda (binding)
           `(define ,@binding)))
        (else
         (##ill-formed-special-form 'infix.define-variable
                                    (##list identifier
                                            type
                                            dims
                                            init)))))

(define-prim (##infix-variable-binding form)
  (if (and (##pair? form)
           (##eq? (##car form) 'infix.define-variable))
    (let ((rest (##cdr form)))
      (if (##pair? rest)
        (let* ((identifier (##car rest))
               (rest (##cdr rest)))
          (if (##pair? rest)
            (let* ((type (##car rest))
                   (rest (##cdr rest)))
              (if (##pair? rest)
                (let* ((dims (##car rest))
                       (rest (##cdr rest)))
                  (if (##pair? rest)
                    (let* ((init (##car rest))
                           (rest (##cdr rest)))
                      (cond ((and (##null? rest)
                                  (##infix-id identifier))
                             =>
                             (lambda (id)
                               `(,id
                                 ,(if (##null? dims)
                                    init
                                    `(infix.make-array ,init ,@dims)))))
                            (else
                             #f)))
                    #f))
                #f))
            #f))
        #f))
    #f))

(define-prim (infix.make-array init . dims)
  (if (##pair? dims)

    (let loop1 ((lst dims) (i 2))
      (let ((dim1 (##car lst)))
        (macro-check-index dim1 i (infix.make-array init . dims)
          (let* ((array (##make-vector dim1 init))
                 (rest (##cdr lst)))
            (if (##pair? rest)
              (let loop2 ((j (##fixnum.- dim1 1)))
                (if (##fixnum.< j 0)
                  array
                  (begin
                    (##vector-set! array j (loop1 rest (##fixnum.+ i 1)))
                    (loop2 (##fixnum.- j 1)))))
              array)))))

    init))

(define-runtime-macro (infix.define-procedure identifier proc)
  `(define ,(##infix-id identifier) ,proc))

(define-runtime-macro (infix.procedure type params stat)
  `(lambda ,(##map (lambda (x) (##infix-id (##car x))) params)
     ,stat))

; There is no predefined semantics for the following infix forms:
;
; (define-runtime-macro (infix.label identifier stat)
;   `(void))
; 
; (define-runtime-macro (infix.goto expr)
;   `(void))
; 
; (define-runtime-macro (infix.switch expr stat)
;   `(void))
; 
; (define-runtime-macro (infix.case expr stat)
;   `(void))
; 
; (define-runtime-macro (infix.break)
;   `(void))
; 
; (define-runtime-macro (infix.continue)
;   `(void))
; 
; (define-runtime-macro (infix.return . expr)
;   `(void))
; 
; (define-runtime-macro (infix.clause expr)
;   `(void))
; 
; (define-runtime-macro (infix.x:-y x y)
;   `(void))
; 
; (define-runtime-macro (infix.cut)
;   `(void))

;------------------------------------------------------------------------------

; Object encoding/decoding.

(define-prim (##object->encoding obj)
  (let* ((hi (##type-cast obj (macro-type-fixnum)))
         (lo (##type obj)))
    (##+ (##* (if (##fixnum.< hi 0) (##- hi ##bignum.2*min-fixnum) hi)
              4)
         lo)))

(define-prim (##encoding->object encoding)
  (let* ((hi (##quotient encoding 4))
         (lo (##modulo encoding 4))
         (x (if (##fixnum? hi) hi (##+ hi ##bignum.2*min-fixnum))))
    (##type-cast x lo)))

(define-prim (##foreign-address ptr)

  (##declare (not interrupts-enabled))

  (let ((addr (##vector-ref ptr 2)))

    ; Beware! "addr" is an arbitrary C pointer.  If the GC ever
    ; gets hold of it, thinking that it is a reference to a Scheme
    ; datum, it is bound to cause trouble.  Fortunately, because
    ; "addr" is dead after the binding of "hi" and "lo", the
    ; compiler will overwrite the register containing "addr" with
    ; "lo".

    (let* ((hi (##type-cast addr (macro-type-fixnum)))
           (lo (##type addr)))

      (##declare (interrupts-enabled))

      (##+ (##* (if (##fixnum.< hi 0) (##- hi ##bignum.2*min-fixnum) hi)
                4)
           lo))))

;==============================================================================
