/* file: "main.c" */

/* Copyright (C) 1994-2002 by Marc Feeley, All Rights Reserved. */

/* This is the driver of the Gambit-C system */

#define ___INCLUDED_FROM_MAIN
#define ___VERSION 40013
#include "gambit.h"

#include "os.h"
#include "setup.h"


___HIDDEN ___SCMOBJ usage_err
   ___P((char **argv,
         int debug_settings),
        (argv,
         debug_settings)
char **argv;
int debug_settings;)
{
  ___setup_params.debug_settings = debug_settings;

  if (___DEBUG_SETTINGS_LEVEL(debug_settings) != 0)
    {
      char *msgs[4];
      msgs[0] = "Usage: ";
      msgs[1] = argv[0];
      msgs[2] =
        " [-:OPTION,OPTION...] ...\n"
        "where OPTION is one of:\n"
        "  mSIZE      set minimum heap size in kilobytes\n"
        "  hSIZE      set maximum heap size in kilobytes\n"
        "  lPERCENT   set heap live ratio after GC in percent\n"
        "  s          set standard Scheme mode\n"
        "  d[OPT...]  set debugging options; OPT is one of:\n"
        "               p/a       treat uncaught exceptions as errors\n"
        "                         (primordial-thread only/all threads)\n"
        "               r/s/q     error handling (create a new REPL/start in\n"
        "                         single-step mode/quit with error status)\n"
        "               i/c/-     select REPL interaction channel\n"
        "                         (ide/console/standard input and output)\n"
        "               0..9      verbosity level\n"
        "  t[OPT...]  set terminal options; OPT is one of:\n"
        "               e/E       enable terminal line-editing (on/off)\n"
        "               n/1/2/4/8 character encoding (native/LATIN1/UCS2/\n"
        "                         UCS4/UTF8)\n"
        "  =DIRECTORY override the setting of GAMBCDIR\n"
        "  +ARG       add argument to the command line before other arguments\n";
      msgs[3] = 0;
      ___display_error (msgs);
    }

  return ___FIXADD(___FIX(___EXIT_CODE_USAGE),___FIX(1));
}


___HIDDEN void skip_string
   ___P((char **start),
        (start)
char **start;)
{
  char c;
  char *p = *start;

  while ((c = *p) != '\0' && c != ',')
    {
      p++;
      if (c == '\\')
        {
          if ((c = *p) == '\0')
            break;
          p++;
        }
    }

  *start = p;
}


___HIDDEN char *extract_string
   ___P((char **start),
        (start)
char **start;)
{
  char c;
  char *p1 = *start;
  char *p2;
  int n = 0;
  char *result;

  while ((c = *p1) != '\0' && c != ',')
    {
      p1++;
      if (c == '\\')
        {
          if ((c = *p1) == '\0')
            break;
          p1++;
        }
      n++;
    }

  p2 = *start;
  *start = p1;

  result = ___CAST(char*,___alloc_mem (n+1));

  if (result != 0)
    {
      p1 = result;
      while ((c = *p2) != '\0' && c != ',')
        {
          p2++;
          if (c == '\\')
            {
              if ((c = *p2) == '\0')
                break;
              p2++;
            }
          *p1++ = c;
        }
      *p1++ = '\0';
    }

  return result;
}


___HIDDEN ___BOOL extend_argv
   ___P((char ***argv,
         int pos,
         int nb_to_add,
         ___BOOL free_old),
        (argv,
         pos,
         nb_to_add,
         free_old)
char ***argv;
int pos;
int nb_to_add;
___BOOL free_old;)
{
  int n = 0;
  char **old_argv = *argv;
  char **new_argv;

  while (old_argv[n++] != 0) ;

  new_argv = ___CAST(char**,___alloc_mem ((n+nb_to_add) * sizeof (char*)));

  if (new_argv == 0)
    return 0;

  while (n >= pos)
    {
      new_argv[n+nb_to_add] = old_argv[n];
      n--;
    }

  while (n >= 0)
    {
      new_argv[n] = old_argv[n];
      n--;
    }

  *argv = new_argv;

  if (free_old)
    ___free_mem (old_argv);

  return 1;
}


___EXP_FUNC(int,___main)
  ___P((int argc,
        char **argv,
        ___mod_or_lnk (*linker)(___global_state_struct*)),
       (argc,
        argv,
        linker)
int argc;
char **argv;
___mod_or_lnk (*linker)();)
{
#define LARGEST_INT (int)((~___CAST(unsigned int,0))>>1)

  char *runtime_options;
  char **current_argv;
  int extra_arg_pos;
  int contract_argv;
  int options_source;
  char *gambcdir;
  long min_heap_len;
  long max_heap_len;
  int live_percent;
  int standard_level;
  int debug_settings;
  int terminal_settings;
  ___SCMOBJ err = ___FIX(___HEAP_OVERFLOW_ERR);
  ___setup_params_struct setup_params;

  /* handle arguments to runtime */

  gambcdir = 0;
  min_heap_len = 0;
  max_heap_len = 0;
  live_percent = 0;
  standard_level = 0;
  debug_settings = ___DEBUG_SETTINGS_INITIAL;
  terminal_settings = ___TERMINAL_SETTINGS_INITIAL;

  runtime_options = argv[1];

  if (runtime_options != 0 &&
      runtime_options[0] == '-' &&
      runtime_options[1] == ':')
    runtime_options += 2;
  else
    runtime_options = 0;

  current_argv = argv;
  extra_arg_pos = 1;
  contract_argv = (runtime_options != 0);

  for (options_source=0; options_source<2; options_source++)
    {
      char *arg;

      if (options_source == 0)
        arg = ___getenv (___GAMBCOPT_ENV_VAR);
      else
        arg = runtime_options;

      if (arg == 0)
        continue;

      do
        {
          char *s = arg++;
          switch (*s)
            {
            case 'm':
            case 'h':
            case 'l':
              {
                int argval = 0;
                while (*arg >= '0' && *arg <= '9')
                  {
                    int n = *arg - '0';
                    if (argval > (LARGEST_INT>>10)/10 ||
                        (argval == (LARGEST_INT>>10)/10 &&
                         n > ((LARGEST_INT>>10)-argval*10)))
                      {
                        err = usage_err (argv, debug_settings);
                        goto after_setup;
                      }
                    argval = argval*10 + n;
                    arg++;
                  }
                if (arg == s+1)
                  {
                    err = usage_err (argv, debug_settings);
                    goto after_setup;
                  }
                else
                  argval = atoi (s+1);
                switch (*s)
                  {
                  case 'm': min_heap_len = argval<<10;
                            break;
                  case 'h': max_heap_len = argval<<10;
                            break;
                  case 'l': live_percent = argval;
                            break;
                  }
                break;
              }

            case 's':
              standard_level = 4;
              break;

            case 'd':
              {
                if (*arg == '\0' || *arg == ',')
                  debug_settings = ___DEBUG_SETTINGS_DEFAULT;
                else
                  while (*arg != '\0' && *arg != ',')
                    {
                      if (*arg >= '0' && *arg <= '9')
                        debug_settings =
                          (debug_settings & ~___DEBUG_SETTINGS_LEVEL_MASK)
                          | ((*arg - '0') << ___DEBUG_SETTINGS_LEVEL_SHIFT);
                      else
                        switch (*arg)
                          {
                          case 'p': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_UNCAUGHT_MASK)
                                      | (___DEBUG_SETTINGS_UNCAUGHT_PRIMORDIAL
                                         << ___DEBUG_SETTINGS_UNCAUGHT_SHIFT);
                                    break;
                          case 'a': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_UNCAUGHT_MASK)
                                      | (___DEBUG_SETTINGS_UNCAUGHT_ALL
                                         << ___DEBUG_SETTINGS_UNCAUGHT_SHIFT);
                                    break;
                          case 'r': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_ERROR_MASK)
                                      | (___DEBUG_SETTINGS_ERROR_REPL
                                         << ___DEBUG_SETTINGS_ERROR_SHIFT);
                                    break;
                          case 's': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_ERROR_MASK)
                                      | (___DEBUG_SETTINGS_ERROR_SINGLE_STEP
                                         << ___DEBUG_SETTINGS_ERROR_SHIFT);
                                    break;
                          case 'q': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_ERROR_MASK)
                                      | (___DEBUG_SETTINGS_ERROR_QUIT
                                         << ___DEBUG_SETTINGS_ERROR_SHIFT);
                                    break;
                          case 'i': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_REPL_MASK)
                                      | (___DEBUG_SETTINGS_REPL_IDE
                                         << ___DEBUG_SETTINGS_REPL_SHIFT);
                                    break;
                          case 'c': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_REPL_MASK)
                                      | (___DEBUG_SETTINGS_REPL_CONSOLE
                                         << ___DEBUG_SETTINGS_REPL_SHIFT);
                                    break;
                          case '-': debug_settings =
                                      (debug_settings
                                       & ~___DEBUG_SETTINGS_REPL_MASK)
                                      | (___DEBUG_SETTINGS_REPL_STDIO
                                         << ___DEBUG_SETTINGS_REPL_SHIFT);
                                    break;
                          default:
                            err = usage_err (argv, debug_settings);
                            goto after_setup;
                          }
                      arg++;
                    }
                break;
              }

            case 't':
              {
#if 0
                if (___TERMINAL_SETTINGS_LEVEL(terminal_settings)
                    == ___TERMINAL_SETTINGS_LEVEL_MASK)
                  terminal_settings = ___TERMINAL_SETTINGS_INITIAL;
#endif

                while (*arg != '\0' && *arg != ',')
                  {
                    switch (*arg++)
                      {
                      case 'e': terminal_settings = 0;
                                break;
                      case 'E': terminal_settings = 0;
                                break;
                      case 'c': terminal_settings = 0;
                                break;
                      case '1': terminal_settings = 0;
                                break;
                      case '2': terminal_settings = 0;
                                break;
                      case '4': terminal_settings = 0;
                                break;
                      case '8': terminal_settings = 0;
                                break;
                      default:
                        err = usage_err (argv, debug_settings);
                        goto after_setup;
                      }
                  }
                break;
              }

            case '=':
              {
                if (gambcdir != 0)
                  ___free_mem (gambcdir);
                gambcdir = extract_string (&arg);
                if (gambcdir == 0)
                  goto after_setup;
                break;
              }

            case '+':
              {
                char *extra_arg;

                if (!extend_argv (&current_argv,
                                  extra_arg_pos,
                                  1 - contract_argv,
                                  current_argv != argv))
                  goto after_setup;

                contract_argv = 0;

                if ((extra_arg = extract_string (&arg)) == 0)
                  goto after_setup;

                current_argv[extra_arg_pos++] = extra_arg;

                break;
              }

            default:
              err = usage_err (argv, debug_settings);
              goto after_setup;
            }
        }

      while (*arg++ == ',') ;

      if (*--arg != '\0')
        {
          err = usage_err (argv, debug_settings);
          goto after_setup;
        }
    }

  if (contract_argv != 0)
    {
      if (!extend_argv (&current_argv, extra_arg_pos, -1, 0))
        goto after_setup;
    }

#ifdef FORCE_MAX_HEAP
  if (max_heap_len == 0 || max_heap_len > FORCE_MAX_HEAP<<10)
    max_heap_len = FORCE_MAX_HEAP<<10;
#endif

  /* Setup program, run it and perform any cleanup necessary. */

  ___setup_params_reset (&setup_params);

  setup_params.version           = ___VERSION;
  setup_params.argv              = current_argv;
  setup_params.min_heap          = min_heap_len;
  setup_params.max_heap          = max_heap_len;
  setup_params.live_percent      = live_percent;
  setup_params.standard_level    = standard_level;
  setup_params.debug_settings    = debug_settings;
  setup_params.terminal_settings = terminal_settings;
  setup_params.gambcdir          = (gambcdir != 0)
                                   ? gambcdir
                                   : ___getenv (___GAMBCDIR_ENV_VAR);
  setup_params.linker            = linker;

  err = ___setup (&setup_params);

 after_setup:

  while (extra_arg_pos > 1)
    ___free_mem (current_argv[--extra_arg_pos]);

  if (current_argv != argv)
    ___free_mem (current_argv);

  if (gambcdir != 0)
    ___free_mem (gambcdir);

  if (err == ___FIX(___NO_ERR))
    {
      ___cleanup ();
      err = ___FIX(___EXIT_CODE_OK);
    }
  else if (err > ___FIX(___NO_ERR))
    err = ___FIXSUB(err,___FIX(1));
  else
    {
      char *msgs[2];
      char msg[___ERR_MAX_LENGTH+1];
      ___err_code_to_string (___INT(err), msg);
      msgs[0] = msg;
      msgs[1] = 0;
      ___display_error (msgs);
      err = ___FIX(___EXIT_CODE_OSERR);
    }

  return ___INT(err);
}
